@@ -7,7 +7,6 @@ package tools
 
 import (
 	_ "github.com/cpuguy83/go-md2man/v2"
-	_ "github.com/onsi/ginkgo/v2/ginkgo"
+	_ "github.com/go-swagger/go-swagger/cmd/swagger"
 	_ "github.com/vbatts/git-validation"
-	_ "golang.org/x/tools/cmd/goimports"
 )
